#!/usr/bin/env python

# Chord Extractor
# Copyright (C) 2021-22  Oliver Holloway
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


from chord_extractor.base import ChordExtractor, ChordChange
import librosa
import vamp
from typing import List
from enum import Enum
import os
import sys
from pkg_resources import resource_filename
import logging

_log = logging.getLogger(__name__)

if not os.getenv('VAMP_PATH') and sys.platform == 'linux' and sys.maxsize > 2 ** 32:
    os.environ['VAMP_PATH'] = os.path.dirname(resource_filename('chord_extractor', '_lib/nnls-chroma.so'))
elif not os.getenv('VAMP_PATH'):
    _log.warning('Please make sure VAMP_PATH is specified pointing to the directory containing the '
                 'appropriate compiled library for Chordino. '
                 'This can be sourced by using the Vamp Plugin Pack installer at '
                 'https://code.soundsoftware.ac.uk/projects/vamp-plugin-pack')


class TuningMode(Enum):
    """Options for Tuning Mode parameter (see Chordino doc)"""
    GLOBAL = 0
    LOCAL = 1


class ChromaNormalization(Enum):
    """Options for Chroma Normalization parameter (see Chordino doc)"""
    NONE = 0
    MAX = 1
    L1 = 2
    L2 = 3


class Chordino(ChordExtractor):
    """
    Class for extracting chords using Chordino (http://www.isophonics.net/nnls-chroma). All parameters are those
    passed to Chordino. You can see further details of what they mean in the
    Chordino -> Parameters section in the above link. The defaults here are the suggested parameter settings for
    when extracting from a generic pop song. See the link for more recommended settings.

    :param use_nnls: Use approximate transcription (NNLS)
    :param roll_on: Spectral roll-on (range: 0 - 5)
    :param tuning_mode: Tuning mode
    :param spectral_whitening: Spectral whitening (range: 0 - 1)
    :param spectral_shape: Spectral shape (range: 0.5 - 0.9)
    :param boost_n_likelihood: Boost likelihood of the N (no chord) label
    :param kwargs: Any other parameters that may become available to the chordino vamp plugin. Param keys are the
     vamp identifier.
    """

    def __init__(self,
                 use_nnls: bool = True,
                 roll_on: float = 1,
                 tuning_mode: TuningMode = TuningMode.GLOBAL,
                 spectral_whitening: float = 1,
                 spectral_shape=0.7,
                 boost_n_likelihood: float = 0.1,
                 **kwargs):
        super().__init__()
        self._params = {
            'useNNLS': int(use_nnls),
            'rollon': roll_on,
            'tuningmode': tuning_mode.value,
            'whitening': spectral_whitening,
            's': spectral_shape,
            'boostn': boost_n_likelihood
        }
        self._params.update(kwargs)

    def extract(self, file: str, **kwargs) -> List[ChordChange]:
        """
        Extract chord changes from a particular file. The file is loaded into librosa, therefore takes sound files
        supported by librosa (which uses audioread and soundfile). This includes .wav, .mp3, .ogg and others.

        :param file: Absolute file path to the relevant file. A file like object is also acceptable.
        :param kwargs: Keyword arguments for librosa.load
         (see https://librosa.org/doc/0.7.0/generated/librosa.core.load.html)
        :return: List of chord changes for the sound file
        """
        data, rate = librosa.load(file, **kwargs)
        _log.info('Submitting {} to Chordino for chord extraction.'.format(file))
        chords = vamp.collect(data, rate, 'nnls-chroma:chordino', parameters=self._params)
        _log.info('Chord extraction for {} complete.'.format(file))
        return [ChordChange(timestamp=float(change['timestamp']),
                            chord=change['label']) for change in chords['list']]



#!/usr/bin/env python

# Chord Extractor
# Copyright (C) 2021-22  Oliver Holloway
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from abc import ABC, abstractmethod
from typing import List, Callable, Optional
import multiprocessing as mp
from .converters import midi_to_wav
import logging
import os
import glob
from .outputs import ChordChange, LabelledChordSequence
import signal


_log = logging.getLogger(__name__)
_tmp_root = os.getenv('EXTRACTOR_TEMP_FILE_PATH', '/tmp/')
_tmp_dir = os.path.join(_tmp_root, 'extractor/')
if not os.path.exists(_tmp_dir):
    os.makedirs(_tmp_dir)


def _convert_to_intermediate_file(instance, path, extract_q, file_count_q):
    file_count_q.put(path)
    intermediate_file = getattr(instance, 'preprocess')(path)
    extract_q.put((intermediate_file, True) if intermediate_file else (path, False))


def clear_conversion_cache():
    """Clear the temporary directory containing sound file conversions."""
    files = glob.glob(os.path.join(_tmp_dir, '*'))
    for f in files:
        os.remove(f)


def _error_cb(e):
    _log.exception(e)
    os.kill(os.getpid(), signal.SIGKILL)


class ChordExtractor(ABC):
    """
    Abstract class for extracting chords from sound files. It also provides functionality for sound file conversion,
    which can be added to by overriding .convert. This is useful as implementations using .extract
    may take only certain file formats, and integrating conversion logic here enables it to be parallelized.
    """

    @abstractmethod
    def extract(self, file: str) -> List[ChordChange]:
        """
        Extract chord changes from a particular file

        :param file: File path to the relevant file (and if possible can accept file like object)
        :return: List of chord changes for the sound file
        """

    def preprocess(self, path: str) -> Optional[str]:
        """
        Run any preprocessing steps based on the location path of the sound file provided. Primarily this is used to
        convert the file at the path, based on its file extension to a file usable by the extract method. However, an
        override of this method can perform any logic that may benefit from multiprocessing available in extract_many.

        In this implementation any midi files are converted to wav files which are placed in a temporary directory
        (conversion cache).

        :param path: Path to the file
        :return: Return file path to output file of conversion if conversion has happened, else None
        """
        ext = os.path.splitext(path)[1]
        if ext in ['.mid', '.midi']:
            return midi_to_wav(path, _tmp_dir)
        return None

    def extract_many(self,
                     files: List[str],
                     callback: Callable[[LabelledChordSequence], None] = None,
                     num_extractors: int = 1,
                     num_preprocessors: int = 1,
                     max_files_in_cache: int = 50,
                     stop_on_error=False) -> List[LabelledChordSequence]:
        """
        Extract chords from a list of files. As opposed to looping over the .extract method, this one gives the option
        to run many extractions in parallel. Furthermore any file conversions that have to be done as a prerequisite
        to extraction can also be done in parallel with the extractions.

        Files can be a mix of different file formats. If a file is of a format that needs to be converted in
        preprocessing, it is the conversion that is passed to a queue for extraction. Otherwise the original file will
        be queued. Any conversions are cached in a temporary folder specified using the environment variable
        EXTRACTOR_TEMP_FILE_PATH (/tmp if not specified). Note, if in any subsequent extract runs, an existing file
        conversion is found in the temporary directory, a new conversion will be skipped and the existing one used.

        :param files: List of paths to files we wish to extract chords for
        :param callback: An optional callable that is called when chords have been extracted from a particular file.
         This takes a tuple with the extraction results and original filepath of the extracted file as argument.
        :param num_extractors: Max number of extraction processes to run in parallel
        :param num_preprocessors: Max number of conversion processes to run in parallel
        :param max_files_in_cache: Limit of number of files for a single extract_many run to have in the temporary file
         cache (for file conversions) at any one time. If 0, there is no limit and no conversions will be deleted,
         otherwise conversions are deleted once extractions are performed (extractions are paused if the file limit
         is reached).
        :param stop_on_error: If True, an error encountered during a single extraction will stop the overall
         extraction process. Be warned, this means that the parent process will be killed. If False, an error will
         cause a None result to be returned in the sequence attribute for a particular LabelledChordSequence result.
        :return: List of tuples, each with the extraction results and id being the file that the extraction was taken
         from.
        """
        m = mp.Manager()
        extract_q = m.Queue()
        file_count_q = m.Queue(maxsize=max_files_in_cache)
        conversion_pool = mp.Pool(processes=num_preprocessors)
        extractor_pool = mp.Pool(processes=num_extractors)
        conversions = [conversion_pool.apply_async(_convert_to_intermediate_file,
                                                   args=(self, file, extract_q, file_count_q))
                       for file in files]
        extractions = []
        while len(extractions) < len(files):
            path, remove_path = extract_q.get()
            extractions.append((path, extractor_pool.apply_async(self._consume,
                                                                 args=(path, file_count_q),
                                                                 kwds={'remove_path': remove_path
                                                                                      and not not max_files_in_cache,
                                                                       'stop_on_error': stop_on_error},
                                                                 callback=callback,
                                                                 error_callback=_error_cb)))

        for c in conversions:
            c.get()
        res = [a[1].get() for a in extractions]
        return res

    def _consume(self, path, file_count_q, remove_path=False, stop_on_error=False) -> LabelledChordSequence:
        res = None
        try:
            res = self.extract(path)
        except Exception as e:
            _log.error('Error has been encountered with extracting chords from {}.'.format(path))
            if stop_on_error:
                raise
            _log.exception(e)
            _log.info('Proceeding to next extraction')

        finally:
            file_count_q.get()
        if remove_path:
            os.remove(path)
        return LabelledChordSequence(id=path, sequence=res)



#!/usr/bin/env python

# Chord Extractor
# Copyright (C) 2021-22  Oliver Holloway
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
import subprocess
import logging

_log = logging.getLogger(__name__)


def midi_to_wav(midi_path: str, wav_to_dir: str):
    """
    Convert midi at given path to wav file and save in specified output directory. This is done using
    Timidity (http://timidity.sourceforge.net/). If an invalid midi file is passed, note that no exception is raised
    and an output file containing invalid wav data.

    :param midi_path: Path to input midi file
    :param wav_to_dir: Path to the output directory
    :return: Path to the new wav file
    """
    base = os.path.basename(midi_path)
    file_name = os.path.splitext(base)[0]
    wav_file = os.path.join(wav_to_dir, file_name)
    if not os.path.isfile(wav_file):
        _log.info('Running timidity on {} to create {}'.format(midi_path, wav_file))
        subprocess.run(['timidity', midi_path, '-Ow', '-o', wav_file])
    else:
        _log.info('Returning already existing temporary wav file {}'.format(wav_file))
    return wav_file



#!/usr/bin/env python

# Chord Extractor
# Copyright (C) 2021-22  Oliver Holloway
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Python library for extracting chord sequences from sound files of multiple formats with the option of leveraging
multiprocessing to source data from many files quickly. The extraction process wraps Chordino
(http://www.isophonics.net/nnls-chroma) but is extensible to easily incorporate additional techniques.

Simple usage::

    from chord_extractor.extractors import Chordino

    chordino = Chordino(roll_on=1)

    # Optional, only if we need to extract from a file that isn't accepted by librosa
    conversion_file_path = chordino.preprocess('/some_path/some_song.mid')

    chords = chordino.extract(conversion_file_path)
    # => [  ChordChange(chord='N', timestamp=0.371519274),
    #       ChordChange(chord='C', timestamp=0.743038548),
    #       ChordChange(chord='Am7b5', timestamp=8.54494331),...]

Extract from many files with multiprocessing::

    from chord_extractor.extractors import Chordino
    from chord_extractor import clear_conversion_cache, LabelledChordSequence

    files_to_extract_from = [
      '/path/file1.mid',
      '/path/file2.wav',
      '/path/file3.mp3',
      '/path/file4.ogg'
    ]

    def save_to_db_cb(results: LabelledChordSequence):
        # Every time one of the files has had chords extracted, receive the chords here
        # along with the name of the original file and then run some logic here, e.g. to
        # save the latest data to DB

    chordino = Chordino(roll_on=1)

    # Optionally clear cache of file conversions (e.g. wav files that have been converted from midi)
    clear_conversion_cache()
    res = chordino.extract_many(files_to_extract_from, callback=save_to_db_cb, num_extractors=2,
                                num_preprocessors=2, max_files_in_cache=10, stop_on_error=False)
    # => LabelledChordSequence(
    #	    id='/tmp/extractor/d8b8ab2f719e8cf40e7ec01abd116d3a',
    #	    sequence=[ChordChange(chord='N', timestamp=0.371519274),
    #	        ChordChange(chord='C', timestamp=0.743038548),
    #	        ChordChange(chord='Am7b5', timestamp=8.54494331),...])
"""

from .base import ChordExtractor, clear_conversion_cache
from .outputs import ChordChange, LabelledChordSequence

__all__ = ['ChordExtractor', 'clear_conversion_cache', 'ChordChange', 'LabelledChordSequence']



#!/usr/bin/env python

# Chord Extractor
# Copyright (C) 2021-22  Oliver Holloway
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from typing import NamedTuple, List, Optional


class ChordChange(NamedTuple):
    """
    Tuple to represent the time point in a sound file at which a chord changes and which chord it changes to.

    The string notation used in the chord string is standard notation (and that used by Chordino). Examples include
    Emaj7, Dm, Cb7, where maj = major, m = minor, b = flat etc. Note that if 'N' is given, this denotes no chord, e.g.
    if a part of the sound file is non-musical.
    """
    chord: str
    timestamp: float


class LabelledChordSequence(NamedTuple):
    """
    Output of chord extractions with identifier, suitable for when running using asynchronous processes. The
    sequence may be None if no result was returned for that particular id.
    """
    id: str
    sequence: Optional[List[ChordChange]]



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

version = {}
with open('chord_extractor/version.py') as fp:
    exec(fp.read(), version)

setuptools.setup(
    name="chord-extractor",
    version=version['__version__'],
    author="Oliver Holloway",
    author_email="oholloway.consulting@gmail.com",
    description="Python library for extracting chords from multiple sound file formats",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ohollo/chord-extractor",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6,<3.10',
    install_requires=[
        'librosa', 'vamp'
    ],
    package_data={'chord_extractor': ['_lib/nnls-chroma.so']}
)